@@ -3,7 +3,7 @@ from datetime import datetime, timedelta
 from typing import Any, Dict, Optional, Union
 
 from pyspark.sql import DataFrame
-from pyspark.sql.functions import lit
+from pyspark.sql.functions import lit, format_string
 from pyspark.sql.types import StructType
 
 from ng_data_pipelines_sdk.aws_interface import AWSInterface
@@ -17,6 +17,7 @@ from ng_data_pipelines_sdk.interfaces import (
     FnKind,
     InputDataFrameParams,
     InputDataFrameParamsDict,
+    Layer,
     OutputDataFrameParams,
     OutputDataFrameParamsDict,
     S3BucketParams,
@@ -162,9 +163,8 @@ class DataFrameManager:
             )
 
             # Prevent pyspark from adding date columns to the DataFrame
-            if not input_df_params.read_date_params.add_all_date_columns:
+            if input_df_params.read_date_params.do_not_include_date_columns:
                 full_base_path = None
-
         else:
             paths = f"{bucket_url}/{input_df_params.dataframe_base_path}"
 
@@ -194,6 +194,28 @@ class DataFrameManager:
 
         if df is None:
             raise EmptyDataFrameException
+        else:
+            # Convert date partitions to string and add leading zeros because PySpark reads them as integers
+            if input_df_params.read_date_params and full_base_path:
+                date_partitions = input_df_params.read_date_params.date_partitions
+                year_col_name = date_partitions.get(DatePartition.YEAR, "year")
+                month_col_name = date_partitions.get(DatePartition.MONTH, "month")
+                day_col_name = date_partitions.get(DatePartition.DAY, "day")
+
+                df = (
+                    df.withColumn(
+                        year_col_name,
+                        format_string("%04d", df[year_col_name]),
+                    )
+                    .withColumn(
+                        month_col_name,
+                        format_string("%02d", df[month_col_name]),
+                    )
+                    .withColumn(
+                        day_col_name,
+                        format_string("%02d", df[day_col_name]),
+                    )
+                )
 
         return df
 
@@ -295,27 +317,41 @@ class DataFrameManager:
         write_base_path_url = f"{bucket_url}/{output_df_params.dataframe_base_path}"
         logger.debug(f"Write base path: {write_base_path_url}")
 
+        write_info = f"Writing DataFrame to base path {write_base_path_url}"
+        if output_df_params.partition_by:
+            write_info += f" with partitions {output_df_params.partition_by}"
+        else:
+            write_info += " without partitions"
+
         df_to_write = df
-        if output_df_params.single_write_date:
-            # Add the date partitions columns to the DataFrame
-            date_partition_path = ""
+        date_partition_path = ""
+        if output_df_params.single_write_date_params:
             for (
                 date_partition,
                 date_partition_name,
-            ) in (
-                output_df_params.single_write_date.single_write_date_partitions.items()
-            ):
-                date_partition_path += f"{date_partition_name}={date_partition.value}/"
-
+            ) in output_df_params.single_write_date_params.single_write_date_partitions.items():
                 formatted_date_part_value = DATE_FORMATTER[date_partition.value](
-                    output_df_params.single_write_date.single_write_date
+                    output_df_params.single_write_date_params.single_write_date
                 )
-
+                # Add the date partition column to the DataFrame
                 df_to_write = df_to_write.withColumn(
                     date_partition_name, lit(formatted_date_part_value)
                 )
+                date_partition_path += (
+                    f"{date_partition_name}={formatted_date_part_value}/"
+                )
 
-            if output_df_params.overwrite:
+            write_info += f". Single write date is being used, so all data will be written to a single path: {write_base_path_url}/{date_partition_path}"
+
+        if output_df_params.overwrite:
+            if (
+                not date_partition_path
+                and output_df_params.dataframe_bucket_params.layer != Layer.ANALYTICS
+            ):
+                logger.warning(
+                    "Overwrite is set to True, but single_write_date was not used and this would delete all data in the base path. This is only allowed in the analytics layer. Skipping overwrite."
+                )
+            else:
                 full_path = f"{write_base_path_url}/{date_partition_path}"
                 logger.info(
                     f"Overwrite is set to True. Deleting existing objects under path {full_path}"
@@ -331,9 +367,8 @@ class DataFrameManager:
 
                 logger.info("Objects deleted successfully.")
 
-        logger.info(
-            f"Writing DataFrame to base path {write_base_path_url}{'' if not output_df_params.partition_by else f' with partitions {output_df_params.partition_by}'}..."
-        )
+        logger.info(f"{write_info}...")
+
         self.spark_manager.write(
             env=output_df_params.dataframe_bucket_params.env,
             df=df_to_write,
@@ -401,24 +436,70 @@ class DataFrameManager:
         return StepParams(**step_config)
 
     def read_input_dataframes_params(
-        self, input_df_params_dict: InputDataFrameParamsDict
+        self,
+        input_df_params_dict: InputDataFrameParamsDict,
     ) -> DataFrameDict:
         """Reads multiple input dataframes based on the provided configuration."""
+
+        def retry_read_dataframe_previous_dates(
+            input_df_params: InputDataFrameParams, max_retries: int = 6
+        ):
+            for retry in range(max_retries + 1):
+                try:
+                    df = self.read_dataframe(
+                        input_df_params=input_df_params, verbose=False
+                    )
+                    return df
+                except EmptyDataFrameException:
+                    if retry == max_retries:
+                        break
+
+                    assert input_df_params.read_date_params is not None
+                    assert isinstance(
+                        input_df_params.read_date_params.read_dates, datetime
+                    )
+
+                    logger.warning(
+                        f"No data found for date {input_df_params.read_date_params.read_dates.strftime('%Y-%m-%d')}. Retrying with previous date..."
+                    )
+                    input_df_params.read_date_params.read_dates -= timedelta(days=1)
+
+            logger.warning(
+                f"Max retries reached. No data found for the last {max_retries + 1} dates."
+            )
+            return None
+
         dfs: DataFrameDict = {}
 
         for df_name, input_df_params in input_df_params_dict.items():
             logger.info(f"Starting read operation for DataFrame '{df_name}'...")
 
-            try:
-                dfs[df_name] = self.read_dataframe(
-                    input_df_params=input_df_params, verbose=False
-                )
-                logger.info(f"Dataframe '{df_name}' read successfully.\n")
-            except EmptyDataFrameException:
-                logger.warning(
-                    f"DataFrame '{df_name}' is empty. No data could be found in the specified path(s). Skipping."
+            if (
+                input_df_params.read_date_params
+                and input_df_params.read_date_params.was_offset_applied
+                and input_df_params.read_date_params.retry_look_window_days > 0
+            ):
+                df = retry_read_dataframe_previous_dates(
+                    input_df_params,
+                    max_retries=input_df_params.read_date_params.retry_look_window_days,
                 )
+            else:
+                try:
+                    df = self.read_dataframe(
+                        input_df_params=input_df_params, verbose=False
+                    )
+                except EmptyDataFrameException:
+                    df = None
 
+            if df is not None:
+                dfs[df_name] = df
+                logger.info(
+                    f"Dataframe '{df_name}' read and added to the dictionary successfully.\n"
+                )
+            else:
+                logger.warning(
+                    f"Dataframe '{df_name}' is empty. It will not be added to the dictionary, and any functions that reference it will fail."
+                )
         return dfs
 
     def _select_transformation_function(self, transform_params: TransformParams):
@@ -501,22 +582,26 @@ class DataFrameManager:
                 df_params.read_date_params.was_offset_applied = False
 
                 if df_params.read_date_params.read_dates == "{{processing_date}}":
-                    df_params.read_date_params.read_dates = processing_date
-
                     offset = df_params.read_date_params.processing_date_offset_days
+
                     if offset and offset != 0:
-                        df_params.read_date_params.read_dates
-                        +timedelta(days=offset)
+                        df_params.read_date_params.read_dates = (
+                            processing_date + timedelta(days=offset)
+                        )
                         df_params.read_date_params.was_offset_applied = True
+                    else:
+                        df_params.read_date_params.read_dates = processing_date
 
         if step_params.output_dataframes_params:
             for _, df_params in step_params.output_dataframes_params.items():
                 if (
-                    df_params.single_write_date
-                    and df_params.single_write_date.single_write_date
+                    df_params.single_write_date_params
+                    and df_params.single_write_date_params.single_write_date
                     == "{{processing_date}}"
                 ):
-                    df_params.single_write_date.single_write_date = processing_date
+                    df_params.single_write_date_params.single_write_date = (
+                        processing_date
+                    )
 
     def process_step(
         self,
@@ -576,7 +661,7 @@ class DataFrameManager:
             input_dfs_dict = self.read_input_dataframes_params(
                 step_params.input_dataframes_params
             )
-            logger.info("Read all Input DataFrames successfully!")
+            logger.info("Finished reading Input DataFrames.")
 
             print("\nPhase 2: Apply Transformations\n" + "-" * 80)
             if step_params.transform_params:
@@ -584,7 +669,7 @@ class DataFrameManager:
                     input_dfs_dict, step_params.transform_params
                 )
                 print("")
-                logger.info("Applied all Transformations successfully!")
+                logger.info("Finished applying transformations to DataFrames.")
             else:
                 print("No transformations to apply. Moving to Phase 3.")
                 dfs_to_output_dict = input_dfs_dict
@@ -594,7 +679,7 @@ class DataFrameManager:
                 self.write_output_dataframes(
                     dfs_to_output_dict, step_params.output_dataframes_params
                 )
-                logger.info("Wrote all Output DataFrames successfully!")
+                logger.info("Finished writing Output DataFrames.")
 
                 if return_output_dfs:
                     return dfs_to_output_dict
