@@ -1,4 +1,5 @@
 import inspect
+from itertools import tee
 import json
 import os
 from datetime import datetime
@@ -252,7 +253,6 @@ class ReadDateParams(BaseModelJsonDumps):
     Attributes:
         read_dates (Union[List[datetime], datetime, Literal["{{processing_date}}"], Literal["{{processing_date_previous}}"]]):
             The dates to be read. It can be a list of datetimes, a single datetime, or special placeholders for processing dates.
-            The default value is "{{processing_date}}".
         processing_date_offset_days (int):
             The offset to be applied to the processing date. It represents the number of days to add or subtract from the processing date.
             The default value is 0.
@@ -273,7 +273,7 @@ class ReadDateParams(BaseModelJsonDumps):
         Sequence[Union[Union[str, datetime], List[Union[str, datetime]]]],
         datetime,
         Literal["{{processing_date}}"],
-    ] = "{{processing_date}}"
+    ]
     processing_date_offset_days: Optional[int] = None
     was_offset_applied: bool = False
     date_partitions: dict[DatePartition, str] = {
@@ -281,7 +281,8 @@ class ReadDateParams(BaseModelJsonDumps):
         DatePartition.MONTH: "month",
         DatePartition.DAY: "day",
     }
-    add_all_date_columns: bool = True
+    do_not_include_date_columns: bool = False
+    retry_look_window_days: int = 7
 
     @model_validator(mode="before")
     def validate_offset_and_read_dates(cls, data):
@@ -296,6 +297,13 @@ class ReadDateParams(BaseModelJsonDumps):
 
         return data
 
+    @field_validator("retry_look_window_days")
+    def ensure_retry_look_window_days_range(cls, v: int) -> int:
+        if not 0 <= v <= 30:
+            raise ValueError("'retry_look_window_days' must be between 0 and 30")
+
+        return v
+
     @field_validator("date_partitions")
     def ensure_all_date_partitions(
         cls, v: dict[DatePartition, str]
@@ -418,9 +426,7 @@ class SingleWriteDateParams(BaseModelJsonDumps):
             The list of partitions to be used for the single write date. Defaults to ["year", "month", "day"].
     """
 
-    single_write_date: Union[datetime, Literal["{{processing_date}}"]] = (
-        "{{processing_date}}"
-    )
+    single_write_date: Union[datetime, Literal["{{processing_date}}"]]
     single_write_date_partitions: dict[DatePartition, str] = {
         DatePartition.YEAR: "year",
         DatePartition.MONTH: "month",
@@ -468,8 +474,7 @@ class OutputDataFrameParams(DataFrameBaseParams):
     dataframe_specific_path: Optional[str] = None
     write_schema_on_s3: bool = False
     overwrite: bool = False
-    dataframe_file_type: FileType = FileType.PARQUET
-    single_write_date: Optional[SingleWriteDateParams] = None
+    single_write_date_params: Optional[SingleWriteDateParams] = None
     partition_by: Optional[List[str]] = None
 
     @model_validator(mode="before")
@@ -489,6 +494,48 @@ class OutputDataFrameParams(DataFrameBaseParams):
 
         return data
 
+    @model_validator(mode="before")
+    def ensure_date_partitions_in_partition_by(cls, data):
+        def is_sublist(smaller, larger):
+            if not smaller:
+                return True
+            if not larger:
+                return False
+
+            it = iter(larger)
+            smaller_it = tee(smaller, 1)[0]
+
+            return any(
+                all(x == y for x, y in zip(smaller_it, it))
+                for _ in range(len(larger) - len(smaller) + 1)
+            )
+
+        single_write_date_params = data.get("single_write_date_params")
+        partition_by = data.get("partition_by")
+
+        if single_write_date_params is None:
+            return data
+
+        if partition_by is None:
+            raise ValueError(
+                "'partition_by' must be provided when using 'single_write_date_params', with at least the date partitions provided in 'single_write_date_params.single_write_date_partitions'."
+            )
+
+        if not is_sublist(
+            single_write_date_params.single_write_date_partitions.values(), partition_by
+        ):
+            raise ValueError(
+                "Date partitions in 'single_write_date_params.single_write_date_partitions' must be a subset (respecting the order) of 'partition_by'."
+            )
+
+        return data
+
+    @field_validator("partition_by")
+    def ensure_unique_partition_by(cls, v: List[str]) -> List[str]:
+        if len(v) != len(set(v)):
+            raise ValueError("'partition_by' must contain unique columns")
+        return v
+
 
 class FnIndirect(BaseModelJsonDumps):
     fn_name: str
